-- This reservoir simulation deck is made available under the Open Database
-- License: http://opendatacommons.org/licenses/odbl/1.0/. Any rights in
-- individual contents of the database are licensed under the Database Contents
-- License: http://opendatacommons.org/licenses/dbcl/1.0/

-- Copyright (C) 2025 Equinor


TCPU

TIMESTEP

FOPR

FGPR

FGIR


FGOR

FWCT

GOPR
  /

GGPR
  /

GGIR
 /

GGOR
 /

WGIR
/

WBHP
/

WOPR
/

ROIP
/

FOIP

FPR

FOPT

GPR
/

GVPR
/

GVIR
/

GWIR
/

FVIR
/

GGIR
/

FGIR
/

FUGINJT

FGSR

GOPGR
/
