/*
  Copyright 2022-2023 SINTEF AS

  This file is part of the Open Porous Media project (OPM).

  OPM is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  OPM is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OPM.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef OPM_GPU_CONSTANTS_HEADER_HPP
#define OPM_GPU_CONSTANTS_HEADER_HPP

#include <cuda_runtime.h>
#include <cusparse.h>

namespace Opm::gpuistl::detail
{

/**
 * @brief The default GPU stream (stream 0)
 */
constexpr cudaStream_t DEFAULT_STREAM = 0;

/**
 * @brief Default matrix ordering for CuSparse operations
 */
constexpr auto CUSPARSE_MATRIX_ORDER = CUSPARSE_DIRECTION_ROW;

} // namespace Opm::gpuistl::detail

#endif
